from __future__ import annotations

import abc
import enum
import logging
import queue
from threading import Thread
from typing import Any
from typing import Literal
from typing import NotRequired
from typing import Self
from typing import Type
from typing import TypedDict
from typing import TypeVar

from ._utils import get_ahk
from ._utils import get_listener
from ._utils import get_logger
from .actions import ActionBase
from .actions import restore_action
from .conditions import ConditionBase
from .conditions import restore_condition
from .voice_listener import Listener

T_Trigger = TypeVar('T_Trigger', bound='TriggerBase')
_trigger_registry: dict[str, Type['TriggerBase']] = {}

logger = get_logger()


def restore_trigger(trigger_data: dict[str, Any]) -> 'TriggerBase':
    trigger_type = trigger_data.get('trigger_type')
    if trigger_type not in _trigger_registry:
        raise ValueError(f'unknown trigger type: {trigger_type!r}')
    klass = _trigger_registry[trigger_type]
    return klass.from_dict(trigger_data)


class TriggerBase:
    def __init__(self) -> None:
        self.actions: list[ActionBase] = []
        self.conditions: list[ConditionBase] = []

    def add_condition(self, condition: ConditionBase) -> Self:
        self.conditions.append(condition)
        return self

    def add_action(self, action: ActionBase) -> Self:
        self.actions.append(action)
        return self

    @classmethod
    def fqn(cls) -> str:
        return f'{cls.__module__}.{cls.__qualname__}'

    @classmethod
    def __init_subclass__(cls: Type[T_Trigger], **kwargs: Any) -> None:
        fqn = cls.fqn()
        assert fqn not in _trigger_registry, f'cannot register class {cls!r} with fqn {fqn!r} which is already in use'
        _trigger_registry[fqn] = cls
        super().__init_subclass__(**kwargs)

    def to_dict(self) -> dict[str, Any]:
        return {
            'trigger_type': self.fqn(),
            'actions': [action.to_dict() for action in self.actions],
            'conditions': [cond.to_dict() for cond in self.conditions],
        }

    @abc.abstractmethod
    def install_hook(self) -> None: ...
    @abc.abstractmethod
    def uninstall_hook(self) -> None: ...

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Self:
        action_type = d.get('trigger_type', None)
        assert action_type == cls.fqn()
        kwargs = d.get('trigger_config', {})
        args: list[Any] | None = None
        variadic_argname: str | None = None
        for argname in kwargs:
            if argname.startswith('*'):
                if args is None:
                    args = kwargs[argname]
                    variadic_argname = argname
                    assert isinstance(args, list)
                else:
                    raise TypeError(f'Multiple variadic properties provided: second variadic: {argname!r}')
        if args is None:
            args = []
        else:
            kwargs.pop(variadic_argname)
        instance = cls(*args, **kwargs)

        for action_data in d.get('actions', []):
            action = restore_action(action_data)
            instance.add_action(action)
        for condition_data in d.get('conditions', []):
            condition = restore_condition(condition_data)
            instance.add_condition(condition)
        return instance

    def check_conditions(self) -> bool:
        return all(cond.check() for cond in self.conditions)

    def on_trigger(self) -> None:
        logger.debug('Checking conditions before performing actions')
        if self.check_conditions():
            logger.info('Conditions check passed. Performing actions')
            for action in self.actions:
                logger.debug('Checking action condition')
                if action.check_conditions():
                    logger.info('Performing action')
                    action.perform()
                else:
                    logger.info('Skipping action due to unmet condition')
            logger.info('Actions')
        else:
            logger.info('Trigger skipped due to unmet condition(s)')


class HotkeyTrigger(TriggerBase):
    class ConfigDict(TypedDict):
        hotkey: str

    def __init__(self, hotkey: str):
        super().__init__()
        self.hotkey = hotkey

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {'trigger_config': {'hotkey': self.hotkey}}

    def install_hook(self) -> None:
        ahk = get_ahk()
        ahk.add_hotkey(self.hotkey, callback=self.on_trigger)
        try:
            ahk.start_hotkeys()
        except Exception:
            pass

    def uninstall_hook(self) -> None:
        ahk = get_ahk()
        ahk.remove_hotkey(self.hotkey)
        try:
            ahk.stop_hotkeys()
        except Exception:
            pass


class JoystickButtonTrigger(HotkeyTrigger):
    class ConfigDict(TypedDict):
        joystick_index: int | str
        joystick_button: str

    def __init__(self, joystick_index: int | str, joystick_button: str):
        self.joystick_index = joystick_index
        self.joystick_button = joystick_button
        super().__init__(hotkey=f'{joystick_index}Joy{joystick_button}')

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {
            'trigger_config': {'joystick_index': self.joystick_index, 'joystick_button': self.joystick_button}
        }


class AxisTriggerMode(enum.IntEnum):
    VALUE_BETWEEN = 1  # Triggers when axis value is within the specified range. Resets when the value falls outside the specified range.
    VALUE_ABOVE = 2  # Triggers when the axis value is greater than the specified value. Resets when the axis value is less than the specified value.
    VALUE_BELOW = 3  # Triggers when the axis value is less than the specified value. Resets when the axis value is greater than the specified value.


class JoystickAxisTrigger(TriggerBase):
    class ConfigDict(TypedDict):
        joystick_index: int
        axis_name: Literal['X', 'Y', 'Z', 'V', 'U', 'R']
        trigger_mode: AxisTriggerMode | Literal[1, 2, 3]
        trigger_value: int | float | tuple[int | float, int | float]
        polling_frequency: NotRequired[int]

    def __init__(
        self,
        joystick_index: int,
        axis_name: str,
        trigger_mode: AxisTriggerMode | int,
        trigger_value: int | float | tuple[int | float, int | float],
        polling_frequency: int = 30,
    ):
        super().__init__()
        self.joystick_index = joystick_index
        self.axis_name = (axis_name,)
        self.trigger_mode = trigger_mode
        self.trigger_value = (trigger_value,)
        self.polling_frequency = polling_frequency

    def _listen_joystick(self) -> None: ...

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {
            'trigger_config': {
                'joystick_index': self.joystick_index,
                'axis_name': self.axis_name,
                'trigger_mode': int(self.trigger_mode),
                'trigger_value': self.trigger_value,
                'polling_frequency': self.polling_frequency,
            },
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Self:
        action_type = d.get('trigger_type', None)
        assert action_type == cls.fqn()
        kwargs = d.get('properties', {})
        trigger_mode_int = kwargs.pop('trigger_mode', None)
        match trigger_mode_int:
            case 1:
                kwargs['trigger_mode'] = AxisTriggerMode.VALUE_BETWEEN
            case 2:
                kwargs['trigger_mode'] = AxisTriggerMode.VALUE_ABOVE
            case 3:
                kwargs['trigger_mode'] = AxisTriggerMode.VALUE_BELOW
            case _:
                raise ValueError(f'invalid trigger mode: {trigger_mode_int}')
        instance = cls(**kwargs)
        for action_data in d.get('actions', []):
            action = ActionBase.from_dict(action_data)
            instance.add_action(action)
        return instance

    def install_hook(self) -> None:
        raise NotImplementedError()

    def uninstall_hook(self) -> None:
        raise NotImplementedError()


class VoiceTrigger(TriggerBase):
    """
    Trigger on spoken voice commands. You can add multiple trigger phrases at once.
    """

    ConfigDict = TypedDict('ConfigDict', {'*trigger_phrases': list[str]})

    def __init__(self, *trigger_phrases: str):
        super().__init__()
        self.trigger_phrases: list[str] = [phrase.lower() for phrase in trigger_phrases]
        self._listener: Listener = get_listener()
        self._match_queue: queue.Queue[str] = self._listener.add_trigger_phrases(self.trigger_phrases)
        self._thread: Thread | None = None

    def _wait_for_trigger(self) -> None:
        while True:
            msg = self._match_queue.get()
            if msg == 'STOP':
                self._match_queue.task_done()
                break
            logging.info(f'received voice trigger for msg: {msg}')
            self.on_trigger()
            self._match_queue.task_done()
        return None

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {'trigger_config': {'*trigger_phrases': self.trigger_phrases}}

    def install_hook(self) -> None:
        assert self._thread is None
        self._thread = Thread(target=self._wait_for_trigger)
        self._thread.start()
        self._listener.start()

    def uninstall_hook(self) -> None:
        assert self._thread is not None
        self._match_queue.put_nowait('STOP')
        self._thread.join()
        self._thread = None
        self._listener.stop()



from __future__ import annotations

import abc
from typing import Any
from typing import NotRequired
from typing import Self
from typing import Type
from typing import TypedDict
from typing import TypeVar
from typing import Unpack

import ahk.exceptions
from ahk import TitleMatchMode

from voice_commander._utils import get_ahk

_condition_registry: dict[str, Type['ConditionBase']] = {}
T_Condition = TypeVar('T_Condition', bound='ConditionBase')


def restore_condition(condition_data: dict[str, Any]) -> 'ConditionBase':
    condition_type = condition_data.get('condition_type')
    if condition_type not in _condition_registry:
        raise ValueError(f'unknown condition type: {condition_type!r}')
    klass = _condition_registry[condition_type]
    return klass.from_dict(condition_data)


class ConditionBase:
    @classmethod
    def fqn(cls) -> str:
        return f'{cls.__module__}.{cls.__qualname__}'

    @classmethod
    def __init_subclass__(cls: Type[T_Condition], **kwargs: Any) -> None:
        fqn = cls.fqn()
        assert fqn not in _condition_registry, f'cannot register class {cls!r} with fqn {fqn!r} which is already in use'
        _condition_registry[fqn] = cls
        super().__init_subclass__(**kwargs)

    @abc.abstractmethod
    def check(self) -> bool: ...

    def to_dict(self) -> dict[str, Any]:
        return {'condition_type': self.fqn()}

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Self:
        condition_type = d.get('condition_type', None)
        assert condition_type == cls.fqn(), f'invalid condition type: {condition_type!r}'
        kwargs = d.get('condition_config', {})
        args: list[Any] | None = None
        variadic_argname: str | None = None
        for argname in kwargs:
            if argname.startswith('*'):
                if args is None:
                    args = kwargs[argname]
                    variadic_argname = argname
                    assert isinstance(args, list)
                else:
                    raise TypeError(f'Multiple variadic properties provided: second variadic: {argname!r}')
        if args is None:
            args = []
        else:
            kwargs.pop(variadic_argname)
        return cls(*args, **kwargs)


class AHKWindowExists(ConditionBase):
    """
    Uses AutoHotkey to check if a given window exists

    Availability: Windows
    """

    class ConfigDict(TypedDict):
        title: NotRequired[str]
        text: NotRequired[str]
        exclude_title: NotRequired[str]
        exclude_text: NotRequired[str]
        title_match_mode: NotRequired[TitleMatchMode]
        detect_hidden_windows: NotRequired[bool]

    def __init__(self, **win_get_kwargs: Unpack[ConfigDict]):
        self.win_get_kwargs = win_get_kwargs

    def check(self) -> bool:
        engine = get_ahk()
        try:
            return engine.win_exists(**self.win_get_kwargs, blocking=False).result()
        except ahk.exceptions.WindowNotFoundException:
            return False

    def to_dict(self) -> dict[str, Any]:
        return {'condition_type': self.fqn(), 'condition_config': self.win_get_kwargs}


class AHKWindowIsActive(ConditionBase):
    """
    Uses AutoHotkey to determine if a window is currently active

    Availability: Windows
    """

    class ConfigDict(TypedDict):
        title: NotRequired[str]
        text: NotRequired[str]
        exclude_title: NotRequired[str]
        exclude_text: NotRequired[str]
        title_match_mode: NotRequired[TitleMatchMode]
        detect_hidden_windows: NotRequired[bool]

    def __init__(self, **win_get_kwargs: Unpack[ConfigDict]):
        self.win_get_kwargs = win_get_kwargs

    def check(self) -> bool:
        engine = get_ahk()
        try:
            return engine.win_is_active(**self.win_get_kwargs, blocking=False).result()
        except ahk.exceptions.WindowNotFoundException:
            return False

    def to_dict(self) -> dict[str, Any]:
        return {'condition_type': self.fqn(), 'condition_config': self.win_get_kwargs}



# Not necessarily intended to be used by users.
# Used to regenerate schema.json file and as local pre-commit hook
from __future__ import annotations

from pathlib import Path


def main() -> None:
    from voice_commander.schema import ProfileSchema

    this_script = Path(__file__)
    schema_file = this_script.parent / 'schema.json'
    expected_schema = ProfileSchema.schema_json(indent=4) + '\n'
    with open(schema_file, encoding='utf-8') as f:
        actual_schema = f.read()

    if expected_schema != actual_schema:
        with open(schema_file, 'w', encoding='utf-8', newline='\n') as f:
            f.write(expected_schema)
        raise SystemExit(1)
    raise SystemExit(0)


if __name__ == '__main__':
    main()



from __future__ import annotations

import json
import os
import pathlib
import sys
import time
import typing
from pathlib import Path
from typing import Any
from typing import Self

import json5

from ._utils import get_listener
from ._utils import get_logger
from .triggers import TriggerBase


logger = get_logger()


def _get_default_profile_path() -> str:
    return os.path.expanduser('~/.voice_commander/profiles')


def load_profile(filepath: str | pathlib.Path) -> 'Profile':
    from .schema import ProfileSchema

    with open(filepath) as f:
        data = json5.load(f)
    return ProfileSchema.parse_obj(data).to_profile()


def load_profile_from_name(profile_name: str) -> 'Profile':
    """
    Loads a profile by name from the default profiles directory ("~/.voice_commander/profiles")

    :param profile_name: the name of the profile
    """
    filename = f'{profile_name}.vcp.json'
    fp = os.path.join(_get_default_profile_path(), filename)
    if not os.path.isfile(fp):
        raise FileNotFoundError(f'Profile file {fp!r} does not exist')
    return load_profile(fp)


class Profile:
    def __init__(self, profile_name: str, triggers: list[TriggerBase] | None = None) -> None:
        self.name = profile_name
        self.triggers: list[TriggerBase] = triggers or []

    def add_trigger(self, trigger: TriggerBase) -> Self:
        self.triggers.append(trigger)
        return self

    def to_dict(self) -> dict[str, Any]:
        from .schema import CURRENT_SCHEMA_VERSION

        return {
            'configuration': {
                'profile_name': self.name,
                'schema_version': CURRENT_SCHEMA_VERSION,
                'triggers': [trigger.to_dict() for trigger in self.triggers],
            }
        }

    def activate(self) -> Self:
        for trigger in self.triggers:
            trigger.install_hook()
        return self

    def deactivate(self) -> None:
        logger.info(f'Deactivating profile {self.name!r}')
        for trigger in self.triggers:
            trigger.uninstall_hook()
        listener = get_listener()
        listener.stop()

    def dump_json(self, file_handle: typing.TextIO, indent: int = 4) -> None:
        json.dump(self.to_dict(), file_handle, indent=indent)

    def dumps_json(self, indent: int = 4) -> str:
        return json.dumps(self.to_dict(), indent=indent)

    def save_json(
        self,
        dirname: Path | str | None = None,
        filename: Path | str | None = None,
        create_dir: bool = True,
        overwrite: bool = False,
    ) -> None:
        if dirname is None:
            dirname = os.path.expanduser('~/.voice_commander/profiles')
        if filename is None:
            filename = f'{self.name}.vcp.json'

        dirname = pathlib.Path(dirname).absolute()
        filename = pathlib.Path(filename)

        if not dirname.exists():
            if create_dir is True:
                os.makedirs(dirname)
            else:
                raise NotADirectoryError(f'dirname {dirname} does not exist.')
        if not dirname.is_dir():
            raise NotADirectoryError(f'dirname {dirname} is not a directory (may be a file)')

        fp = dirname / filename
        if fp.exists():
            if fp.is_dir():
                raise IsADirectoryError(f'{fp} already exists and is a directory, not a file')
            if overwrite is False:
                raise FileExistsError('File already exists but overwrite parameter is False')
        with open(fp, 'w') as outfile:
            self.dump_json(outfile)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        triggers_data = data.get('triggers', [])
        triggers: list[TriggerBase] = []
        for trigger_data in triggers_data:
            t = TriggerBase.from_dict(trigger_data.get('trigger_config', {}))
            triggers.append(t)
        name = data['profile_name']
        return cls(profile_name=name, triggers=triggers)

    def __enter__(self) -> Self:
        self.activate()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.deactivate()

    def run(self) -> None:
        print(f'Running profile {self.name!r}. Press ctrl+c to stop', flush=True, file=sys.stderr)
        with self:
            try:
                while True:
                    time.sleep(0.5)
            except KeyboardInterrupt:
                print('KeyboardInterrupt received. Stopping.', file=sys.stderr, flush=True)
        return None



from __future__ import annotations

import logging
import queue
import time
import warnings
from collections.abc import Sequence
from threading import Lock
from threading import Thread
from typing import TYPE_CHECKING

import speech_recognition as sr  # type: ignore[import-untyped]
from speech_recognition import AudioData
from thefuzz import process  # type: ignore[import-untyped]

from voice_commander._utils import get_logger

logger = get_logger()


class Listener:
    def __init__(self) -> None:
        self._listen_threads: int = 3
        self._listen_timeout: int = 5
        self._listening: bool = False
        self._running: bool = False
        self._mic_lock = Lock()
        self._reporting_lock = Lock()
        self._recognizer_name = 'google'
        self._recognizer = sr.Recognizer()
        self._listener_threads: list[Thread] = []
        self._reporting_queues: dict[str, queue.Queue[str]] = {}
        self._match_threshold: int = 50

    def add_trigger_phrases(self, phrases: Sequence[str]) -> queue.Queue[str]:
        q: queue.Queue[str] = queue.Queue()
        for phrase in phrases:
            phrase = phrase.lower()
            if phrase in self._reporting_queues:
                raise ValueError(f'Phrase {phrase!r} already registered')
            with self._reporting_lock:
                self._reporting_queues[phrase] = q
        return q

    def remove_trigger_phrase(self, phrase: str) -> None:
        if phrase not in self._reporting_queues:
            raise ValueError('Phrase not registered')
        self._reporting_queues.pop(phrase)
        return None

    def _match_command(self, value: str) -> str | None:
        item, ratio = process.extractOne(value, self._reporting_queues.keys())
        if ratio >= self._match_threshold:
            if TYPE_CHECKING:
                assert isinstance(item, str)
            return item
        else:
            return None

    def start(self) -> None:
        if self._running:
            return
        self._running = True
        self._listening = True
        for _ in range(self._listen_threads):
            t = Thread(target=self._listen)
            t.start()
            self._listener_threads.append(t)

    def stop(self) -> None:
        if self._running or self._listening:
            logger.info('Stopping voice listener (this may take a few seconds)...')
            self._running = False
            self._listening = False
            while self._listener_threads:
                t = self._listener_threads.pop()
                t.join()
            logger.info('Voice listener stopped')
        else:
            logger.debug('Stop called on already-stopped listener')

    def start_listening(self) -> None:
        self._listening = True

    def stop_listening(self) -> None:
        self._listening = False

    def _analyze(self, audio: AudioData) -> str | None:
        recognizer_func = getattr(self._recognizer, f'recognize_{self._recognizer_name}', None)
        if recognizer_func is None:
            warnings.warn(f'Unknown recognizer name {self._recognizer_name!r}. Falling back to google recognizer')
            recognizer_func = self._recognizer.recognize_google
        try:
            value = recognizer_func(audio)
            if TYPE_CHECKING:
                assert isinstance(value, str)
            logging.info('Recognized text: "{}"'.format(value))
            return value
        except sr.UnknownValueError as e:
            logging.debug('Issue recognizing audio; {}'.format(e))
        except sr.RequestError as e:
            msg = "Couldn't request results from Google Speech Recognition service; {0}".format(e)
            logging.warning(msg)
        return None

    def _listen(self) -> None:
        logging.debug('LISTENER STARTED')
        while True:
            if not self._listening:
                logging.debug('Listening is currently disabled. Waiting one second.')
                time.sleep(1)
                continue
            if not self._running:
                break
            try:
                with self._mic_lock:
                    if not self._running or not self._listening:
                        break
                    with sr.Microphone() as source:
                        audio = self._recognizer.listen(source, timeout=self._listen_timeout)
                logging.debug('Starting analysis')
                value = self._analyze(audio)
                if value is None:
                    continue
                logging.debug('Attempting to match audio to known phrase')
                matching_phrase = self._match_command(value)
                if matching_phrase is not None:
                    with self._reporting_lock:
                        q = self._reporting_queues[matching_phrase]
                        q.put(value)
            except sr.WaitTimeoutError:
                logging.debug('Timed out waiting for audio')
            except Exception as e:
                logging.error(e, exc_info=True)
            if not self._running:
                break
        logging.debug('LISTENER STOPPED')



from __future__ import annotations

from typing import Self

from ._utils import get_listener
from .profile import Profile


class ApplicationState:
    def __init__(self) -> None:
        self._active_profile_name: str | None = None
        self._all_profiles: dict[str, Profile] = {'default': Profile('default')}
        self._listener = get_listener()

    @property
    def active_profile_name(self) -> str | None:
        return self._active_profile_name

    @active_profile_name.setter
    def active_profile_name(self, new_profile_name: str | None) -> None:
        if self._active_profile_name is not None:
            active_profile = self._all_profiles[self._active_profile_name]
            active_profile.deactivate()
        self._active_profile_name = new_profile_name
        if new_profile_name is not None:
            new_profile = self._all_profiles[new_profile_name]
            new_profile.activate()

    @property
    def active_profile(self) -> None | Profile:
        if self._active_profile_name is not None:
            return self._all_profiles.get(self._active_profile_name, None)
        else:
            return None

    def add_profile(self, profile_name: str, profile: Profile) -> Self:
        assert profile_name not in self._all_profiles, 'Profile name already exists'
        self._all_profiles[profile_name] = profile
        return self

    def delete_profile(self, profile_name: str) -> None:
        assert profile_name in self._all_profiles, 'Profile name does not exist'
        del self._all_profiles[profile_name]

    @property
    def available_profile_names(self) -> list[str]:
        return list(self._all_profiles)



from __future__ import annotations

import abc
import os.path
import time
import warnings
from typing import Any
from typing import NotRequired
from typing import Self
from typing import Type
from typing import TypedDict
from typing import TypeVar
from typing import Unpack

from ahk import TitleMatchMode

from ._utils import get_ahk
from ._utils import get_logger
from .conditions import ConditionBase
from .conditions import restore_condition

winsound: Any = None
try:
    import winsound
except ImportError:
    pass


_action_registry: dict[str, Type['ActionBase']] = {}
T_Action = TypeVar('T_Action', bound='ActionBase')

logger = get_logger()


def restore_action(action_data: dict[str, Any]) -> 'ActionBase':
    action_type = action_data.get('action_type')
    if action_type not in _action_registry:
        raise ValueError(f'unknown action {action_type!r}')
    klass = _action_registry[action_type]
    return klass.from_dict(action_data)


class ActionBase:
    def __init__(self) -> None:
        self.conditions: list[ConditionBase] = []

    @classmethod
    def fqn(cls) -> str:
        return f'{cls.__module__}.{cls.__qualname__}'

    @classmethod
    def __init_subclass__(cls: Type[T_Action], **kwargs: Any) -> None:
        fqn = cls.fqn()
        assert fqn not in _action_registry, f'cannot register class {cls!r} with fqn {fqn!r} which is already in use'
        _action_registry[fqn] = cls
        super().__init_subclass__(**kwargs)

    def add_condition(self, condition: ConditionBase) -> Self:
        self.conditions.append(condition)
        return self

    @abc.abstractmethod
    def perform(self) -> None: ...

    def to_dict(self) -> dict[str, Any]:
        return {'action_type': self.fqn(), 'conditions': [cond.to_dict() for cond in self.conditions]}

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Self:
        # TODO: support variadic parameters, like in TriggerBase.from_dict
        action_type = d.get('action_type', None)
        assert action_type == cls.fqn(), f'action type {action_type!r} was unexpected. Was expecting {cls.fqn()!r}'
        kwargs = d.get('action_config', {})
        args: list[Any] | None = None
        variadic_argname: str | None = None
        for argname in kwargs:
            if argname.startswith('*'):
                if args is None:
                    args = kwargs[argname]
                    variadic_argname = argname
                    assert isinstance(args, list)
                else:
                    raise TypeError(f'Multiple variadic properties provided: second variadic: {argname!r}')
        if args is None:
            args = []
        else:
            kwargs.pop(variadic_argname)

        instance = cls(*args, **kwargs)
        for condition_data in d.get('conditions', []):
            condition = restore_condition(condition_data)
            instance.add_condition(condition)
        return instance

    def check_conditions(self) -> bool:
        return all(cond.check() for cond in self.conditions)


class AHKSendAction(ActionBase):
    """
    Send inputs using AutoHotkey
    """

    class ConfigDict(TypedDict):
        send_string: str

    def __init__(self, send_string: str):
        """
        :param send_string: the keys to send - uses AutoHotkey's `Send <https://www.autohotkey.com/docs/v1/lib/Send.htm>`_ function
        """
        self.send_string: str = send_string
        super().__init__()

    def perform(self) -> None:
        engine = get_ahk()
        engine.send(self.send_string)

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {'action_config': {'send_string': self.send_string}}


class AHKPressAction(ActionBase):
    """
    Press (and release) a single key.
    Adds small delay between key down and key up, useful for making sure games register inputs.
    """

    class ConfigDict(TypedDict):
        key: str

    def __init__(self, key: str):
        """

        :param key: the key to press. Do not include braces for special keys.
        """
        self.key: str = key
        super().__init__()

    def perform(self) -> None:
        engine = get_ahk()
        engine.send(f'{{Blind}}{{{self.key} DOWN}}', blocking=False)
        time.sleep(0.1)
        engine.send(f'{{Blind}}{{{self.key} UP}}', blocking=False)
        time.sleep(0.1)

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {'action_config': {'key': self.key}}


class WinSoundAction(ActionBase):
    """
    Play a sound (wav file)
    """

    class ConfigDict(TypedDict):
        sound_file_path: str

    def __init__(self, sound_file_path: str):
        """

        :param sound_file_path: the path to the sound file. Must be .wav format.
        """
        if winsound is None:
            warnings.warn(
                message='winsound module is not available. WinSoundAction will fail when triggered.',
                category=RuntimeWarning,
                stacklevel=2,
            )
        sound_file_path = os.path.abspath(sound_file_path)
        self.sound_file_path: str = sound_file_path
        if winsound is None:
            self._flags = None
        else:
            self._flags = winsound.SND_FILENAME | winsound.SND_ASYNC | winsound.SND_NOSTOP
        super().__init__()

    def perform(self) -> None:
        if winsound is None:
            raise RuntimeError('winsound module is not available')
        winsound.PlaySound(self.sound_file_path, self._flags)

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {'action_config': {'sound_file_path': self.sound_file_path}}


class AHKMakeWindowActiveAction(ActionBase):
    """
    Make a window active, ensuring it has focus
    """

    class ConfigDict(TypedDict):
        title: NotRequired[str]
        text: NotRequired[str]
        exclude_title: NotRequired[str]
        exclude_text: NotRequired[str]
        title_match_mode: NotRequired[TitleMatchMode]
        detect_hidden_windows: NotRequired[bool]

    def __init__(self, **win_get_kwargs: Unpack[ConfigDict]):
        """

        :param win_get_kwargs:
        """
        self._win_get_kwargs = win_get_kwargs
        super().__init__()

    def perform(self) -> None:
        engine = get_ahk()
        win = engine.win_get(**self._win_get_kwargs)
        assert win is not None, f'Could not find window with parameters {self._win_get_kwargs!r}'
        win.activate()

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {'action_config': self._win_get_kwargs}


class PauseAction(ActionBase):
    """
    Sleeps for specified time
    """

    class ConfigDict(TypedDict):
        seconds: int | float

    def __init__(self, seconds: float | int):
        """

        :param seconds: time in seconds to sleep
        """
        self.seconds: float | int = seconds
        super().__init__()

    def perform(self) -> None:
        time.sleep(self.seconds)

    def to_dict(self) -> dict[str, Any]:
        return super().to_dict() | {'action_config': {'seconds': self.seconds}}



from __future__ import annotations

import argparse
import logging
import sys

from voice_commander._utils import get_logger
from voice_commander.profile import load_profile
from voice_commander.profile import load_profile_from_name

logger = get_logger()

if __name__ == '__main__':
    parser = argparse.ArgumentParser('voice_commander')
    subparsers = parser.add_subparsers(help='sub-commands', dest='subcommand')
    run_profile_subparser = subparsers.add_parser('run_profile', help='Run a profile from a file')
    profile_group = run_profile_subparser.add_mutually_exclusive_group()
    profile_group.add_argument(
        '--profile-name', help='The name of the profile. Looks in the default profiles directory'
    )
    profile_group.add_argument('--profile-file', help='The absolute path to the profile file')
    run_profile_subparser.add_argument('--log-level', choices=('DEBUG', 'INFO', 'WARNING', 'ERROR'), default='INFO')
    args = parser.parse_args()
    if args.subcommand is None:
        parser.print_help()
        raise SystemExit(1)
    if args.subcommand == 'run_profile':
        log_level = getattr(logging, args.log_level, logging.INFO)
        logger.setLevel(log_level)
        if args.profile_name is None and args.profile_file is None:
            print('ERROR: Profile must be provided', file=sys.stderr, flush=True)
            run_profile_subparser.print_help()
            raise SystemExit(1)
        if args.profile_name is not None:
            profile = load_profile_from_name(args.profile_name)
        else:
            profile = load_profile(args.profile_file)
        profile.run()
        raise SystemExit(0)



from __future__ import annotations

import logging
import time
from collections.abc import Callable
from typing import Any
from typing import Union

from ahk import AHK
from ahk.directives import NoTrayIcon


_global_ahk: Union[None, AHK[Any]] = None
_global_listener: Union[None, Listener] = None


def get_logger() -> logging.Logger:
    return logging.getLogger('voice_commander')


def get_ahk(use_global: bool = True) -> AHK[Any]:
    global _global_ahk
    if use_global:
        if _global_ahk is None:
            _global_ahk = AHK(directives=[NoTrayIcon(apply_to_hotkeys_process=True)])
        return _global_ahk
    return AHK(directives=[NoTrayIcon(apply_to_hotkeys_process=True)])


def get_listener() -> Listener:
    global _global_listener
    if _global_listener is None:
        _global_listener = Listener()
    return _global_listener


def get_joystick_button() -> tuple[str, str]:
    """
    Wait for any joystick button to be pressed, return a tuple the AHK index of the
     joystick and the button's name (as strings)

    NOTE: POV hats do not count as buttons. Use get_joystick_axis_or_pov for waiting on POV hat inputs.
    """
    ahk = get_ahk(use_global=False)
    import queue

    q: queue.Queue[tuple[str | int, int]] = queue.Queue()

    def make_callback(index: str | int, button: int) -> Callable[..., None]:
        def inner() -> None:
            q.put((index, button))

        return inner

    for button in range(1, 33):
        ahk.add_hotkey(f'Joy{button}', callback=make_callback('', button))

    for joystick in range(1, 17):
        for button in range(1, 33):
            ahk.add_hotkey(f'{joystick}Joy{button}', callback=make_callback(joystick, button))
    try:
        ahk.start_hotkeys()
        index, button = q.get()
        return str(index), str(button)
    finally:
        ahk.stop_hotkeys()
        del ahk


def _get_all_joystick_values(ahk: AHK[Any]) -> dict[str, dict[str, int | float]]:
    joystick_initial_values = {}
    joystick: str | int
    for joystick in range(0, 17):
        supported_axes = 'XY'
        if joystick == 0:
            joystick = ''
        info_resp = ahk.key_state(f'{joystick}JoyInfo')
        if not info_resp:
            continue
        assert isinstance(info_resp, str)
        for additional_axis in 'ZRUV':
            if additional_axis in info_resp:
                supported_axes += additional_axis
        initial_values = {}
        for axis in supported_axes:
            initial_value = ahk.key_state(f'{joystick}Joy{axis}')
            if isinstance(initial_value, (int, float)):
                initial_values[axis] = initial_value
            else:
                continue
        if 'P' in info_resp:
            pov_value = ahk.key_state(f'{joystick}JoyPOV')
            if isinstance(pov_value, (int, float)):
                initial_values['POV'] = pov_value
        joystick_initial_values[str(joystick)] = initial_values
    return joystick_initial_values


def get_joystick_axis_or_pov() -> tuple[str, str, int | float]:
    ahk = get_ahk(use_global=False)
    # XXX: for some reason, we have to prime this, as some axes report 100 on first poll always
    _get_all_joystick_values(ahk)
    time.sleep(0.1)

    joystick_initial_values = _get_all_joystick_values(ahk)
    while True:
        current_values = _get_all_joystick_values(ahk)
        for joystick, joystick_values in current_values.items():
            initial_values = joystick_initial_values.get(joystick)
            if not initial_values:
                continue
            for axis in joystick_values:
                current_value = joystick_values[axis]
                initial_value = initial_values[axis]
                if abs(current_value - initial_value) > 30 or (axis == 'POV' and current_value != initial_value):
                    return joystick, axis, current_value
        time.sleep(0.1)


from .voice_listener import Listener  # noqa



from __future__ import annotations

from pydantic import BaseModel

from ..profile import Profile
from .v0 import ProfileConfig as ProfileConfigV0

CURRENT_SCHEMA_VERSION = '0'


class ProfileSchema(BaseModel):
    configuration: ProfileConfigV0

    def to_profile(self) -> Profile:
        return self.configuration.to_profile()


__all__ = ['ProfileSchema', 'CURRENT_SCHEMA_VERSION']



from __future__ import annotations

from typing import Any
from typing import Literal
from typing import TypeAlias
from typing import Union

import pydantic
from pydantic import BaseModel
from pydantic import field_validator

from voice_commander.actions import _action_registry
from voice_commander.actions import ActionBase
from voice_commander.actions import AHKMakeWindowActiveAction
from voice_commander.actions import AHKPressAction
from voice_commander.actions import AHKSendAction
from voice_commander.actions import PauseAction
from voice_commander.actions import restore_action
from voice_commander.conditions import AHKWindowExists
from voice_commander.conditions import AHKWindowIsActive
from voice_commander.conditions import ConditionBase
from voice_commander.conditions import restore_condition
from voice_commander.profile import Profile
from voice_commander.triggers import _trigger_registry
from voice_commander.triggers import HotkeyTrigger
from voice_commander.triggers import JoystickAxisTrigger
from voice_commander.triggers import JoystickButtonTrigger
from voice_commander.triggers import restore_trigger
from voice_commander.triggers import TriggerBase
from voice_commander.triggers import VoiceTrigger


class ConditionSchema(BaseModel):
    condition_type: str
    condition_config: dict[str, Any]

    def to_condition(self) -> ConditionBase:
        return restore_condition(self.dict())


class AHKWindowExistsConditionSchema(BaseModel):
    condition_type: Literal['voice_commander.conditions.AHKWindowExists']
    condition_config: AHKWindowExists.ConfigDict

    def to_condition(self) -> AHKWindowExists:
        return AHKWindowExists.from_dict(self.dict())


class AHKWindowIsActiveConditionSchema(BaseModel):
    condition_type: Literal['voice_commander.conditions.AHKWindowIsActive']
    condition_config: AHKWindowIsActive.ConfigDict

    def to_condition(self) -> AHKWindowIsActive:
        return AHKWindowIsActive.from_dict(self.dict())


T_Conditions: TypeAlias = Union[ConditionSchema, AHKWindowExistsConditionSchema, AHKWindowIsActiveConditionSchema]


class AHKSendActionSchema(BaseModel):
    action_type: Literal['voice_commander.actions.AHKSendAction']
    action_config: AHKSendAction.ConfigDict
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    def to_action(self) -> AHKSendAction:
        return AHKSendAction.from_dict(self.dict())


class AHKPressActionSchema(BaseModel):
    action_type: Literal['voice_commander.actions.AHKPressAction']
    action_config: AHKPressAction.ConfigDict
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    def to_action(self) -> AHKPressAction:
        return AHKPressAction.from_dict(self.dict())


class AHKMakeWindowActiveActionSchema(BaseModel):
    action_type: Literal['voice_commander.actions.AHKMakeWindowActiveAction']
    action_config: AHKMakeWindowActiveAction.ConfigDict
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    def to_action(self) -> AHKMakeWindowActiveAction:
        return AHKMakeWindowActiveAction.from_dict(self.dict())


class PauseActionSchema(BaseModel):
    action_type: Literal['voice_commander.actions.PauseAction']
    action_config: PauseAction.ConfigDict
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    def to_action(self) -> PauseAction:
        return PauseAction.from_dict(self.dict())


class ActionSchema(BaseModel):
    action_type: str
    action_config: dict[str, Any]
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    @field_validator('action_type')
    @classmethod
    def action_type_must_exist(cls, v: str) -> str:
        if v not in _action_registry:
            raise ValueError(f'Action type {v!r} does not exist. Must be one of {list(_action_registry.keys())!r}')
        return v

    def to_action(self) -> ActionBase:
        return restore_action(self.dict())


T_Actions: TypeAlias = Union[
    AHKSendActionSchema,
    AHKPressActionSchema,
    AHKMakeWindowActiveActionSchema,
    PauseActionSchema,
    ActionSchema,
]


class HotkeyTriggerSchema(BaseModel):
    trigger_type: Literal['voice_commander.triggers.HotkeyTrigger']
    trigger_config: HotkeyTrigger.ConfigDict
    actions: list[T_Actions]
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    def to_trigger(self) -> HotkeyTrigger:
        return HotkeyTrigger.from_dict(self.dict())


class JoystickAxisTriggerSchema(BaseModel):
    trigger_type: Literal['voice_commander.triggers.JoystickAxisTrigger']
    trigger_config: JoystickAxisTrigger.ConfigDict
    actions: list[T_Actions]
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    def to_trigger(self) -> JoystickAxisTrigger:
        return JoystickAxisTrigger.from_dict(self.dict())


class JoystickButtonTriggerSchema(BaseModel):
    trigger_type: Literal['voice_commander.triggers.JoystickButtonTrigger']
    trigger_config: JoystickButtonTrigger.ConfigDict
    actions: list[T_Actions]
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    def to_trigger(self) -> JoystickButtonTrigger:
        return JoystickButtonTrigger.from_dict(self.dict())


class VoiceTriggerSchema(BaseModel):
    trigger_type: Literal['voice_commander.triggers.VoiceTrigger']
    trigger_config: VoiceTrigger.ConfigDict
    actions: list[T_Actions]
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    def to_trigger(self) -> VoiceTrigger:
        return VoiceTrigger.from_dict(self.dict())


class TriggerSchema(BaseModel):
    trigger_type: str
    trigger_config: dict[str, Any]
    actions: list[T_Actions]
    conditions: list[T_Conditions] = pydantic.Field(default_factory=list)

    @field_validator('trigger_type')
    @classmethod
    def trigger_type_must_exist(cls, v: str) -> str:
        if v not in _trigger_registry:
            raise ValueError(f'Trigger type {v!r} does not exist. Must be one of {list(_trigger_registry.keys())!r}')
        return v

    def to_trigger(self) -> TriggerBase:
        return restore_trigger(self.dict())


T_Triggers: TypeAlias = Union[
    HotkeyTriggerSchema,
    JoystickAxisTriggerSchema,
    JoystickButtonTriggerSchema,
    VoiceTriggerSchema,
    TriggerSchema,
]


class ProfileConfig(BaseModel):
    schema_version: Literal['0'] = '0'
    profile_name: str
    triggers: list[T_Triggers]

    def to_profile(self) -> Profile:
        p = Profile(profile_name=self.profile_name)
        for trigger_def in self.triggers:
            p.add_trigger(trigger_def.to_trigger())
        return p
